/**
 * A unique symbol to identify a model.
 *
 * Converts a model to unwrapped value.
 */
export declare const toModelValue: unique symbol;
/**
 * Since there is no reference or pointer of primitive type in JavaScript,
 *  this is a wrapper of primitive type to make it possible to pass by reference.
 */
export interface JustModel<T> {
    /**
     * The model value.
     */
    value: T;
    /**
     * Convert to unwrapped value.
     */
    [toModelValue](): T;
    /**
     * Convert to primitive value.
     */
    [Symbol.toPrimitive](): T;
}
/**
 * A utility type.
 *
 * Commonly used in component function parameters which accept both primitive type and model.
 */
export type Model<T> = T | JustModel<T>;
/**
 * Create a model.
 *
 * @param v The initial value.
 * @returns The created model.
 */
export declare function model<T>(v: T): JustModel<T>;
/**
 * Check whether a value is a model.
 *
 * @param m The value to check.
 * @returns Whether the value is a model.
 */
export declare function isModel<T>(m: Model<T>): m is JustModel<T>;
/**
 * Update the value of a model.
 *
 * **Warning**: This function is dangerous. It will mutate the model value without triggering an `UPDATE` call.
 *
 * @param m The model.
 * @param v The new value.
 * @returns Whether the value is changed.
 */
export declare function dangerously_updateModel<T>(m: Model<T>, v: T): boolean;
/**
 * Get the unwrapped value.
 *
 * @param v Maybe a model.
 * @returns The unwrapped value.
 */
export declare function unwrap<T>(v: Model<T>): T;
//# sourceMappingURL=model.d.ts.map